from __future__ import annotations
from openpyxl.styles import Font, Alignment, Border, Side
from typing import Iterator

from ..pipeline_contracts import Context, Bundle, LineType, DataLine, SubstanceResult, ReactionResult, Style, Cell, Block

from functools import lru_cache


# ======== styles ========
BLUE = "00008B"
RED = "FF0000"
BLACK = "000000"
DARK_GRAY = "1A1A1A"

thin = Side(border_style="thin", color=BLACK)

DEFAULT_FONT = "Times New Roman"
NORMAL_FONT   = "SimSun"

FONT_SIZE = 11
BIG_FONT_SIZE = 14

# works for Times New Roman
WIDTH_SCALE = FONT_SIZE / 11 * 1.04
WIDTH_EXTRA = 0.71

def col_width_from_str(text:str):
    english_count = sum(1 for ch in text if ord(ch) < 128)
    return (len(text) * 2 - english_count) * WIDTH_SCALE + WIDTH_EXTRA
# def col_width(char_count:int):
#     return char_count * WIDTH_SCALE + WIDTH_EXTRA
ROW_HEADER_WIDTH = 15
NUMBER_WIDTH = 8
EQ_WIDTH = 6

title_style = Style(
        Font(
            name=DEFAULT_FONT,
            size=BIG_FONT_SIZE,
            bold=True,
            color=BLUE,
            ),
        Alignment(
            horizontal="left",
            vertical="center",
            indent=0
            ),
        Border(left=None,right=None,top=None,bottom=None),
        )
row_header_style = title_style

col_header_style = Style(
        Font(
            name=DEFAULT_FONT,
            size=FONT_SIZE,
            bold=True,
            color=BLUE,
            ),
        Alignment(
            horizontal="center",
            vertical="center",
            indent=0,
            shrink_to_fit=True,
            ),
        Border(left=None,right=None,top=None,bottom=None),
        )

table_body_style = Style(
        Font(
            name=DEFAULT_FONT,
            size=FONT_SIZE,
            bold=False,
            color=BLACK,
            ),
        Alignment(
            horizontal="center",
            vertical="center",
            shrink_to_fit=True,
            indent=0
            ),
        Border(left=thin,right=thin,top=thin,bottom=thin),
        )

normal_style = Style(
        Font(
            name=NORMAL_FONT,
            size=FONT_SIZE,
            bold=False,
            color=DARK_GRAY,
            ),
        Alignment(
            horizontal="left",
            vertical="center",
            indent=1
            ),
        Border(left=None,right=None,top=None,bottom=None),
        )

row_header_column = (
       Cell(value="Equation", style=row_header_style),
       Cell(value="Mol. Wt.", style=row_header_style),
       Cell(value="Mol. Stoich.", style=row_header_style),
       Cell(value="Mass (g)", style=row_header_style),
        ) 

null_column = (
       Cell(value="", style=col_header_style),
       Cell(value="", style=table_body_style),
       Cell(value="", style=table_body_style),
       Cell(value="", style=table_body_style),
        ) 

eq_column = (
       Cell(value="  =  ", style=col_header_style),
       Cell(value="", style=table_body_style),
       Cell(value="", style=table_body_style),
       Cell(value="", style=table_body_style),
        ) 


def substance_column(result:SubstanceResult)->tuple[Cell]:
    if result.info:
        header = f"{result.formula}:{result.info}"
    else:
        header = result.formula
    return (
                Cell(header,"@",col_header_style),
                Cell(float(result.mole_mass),"0.#####",table_body_style),
                Cell(int(result.stoich_in_eq),"0",table_body_style),
                Cell(float(result.mass),"0.0000",table_body_style),
                )


# ======== Stage ========
class FormatStage:
    def __init__(self):
        self.reaction_rows = []
        self.null_start_cols = []
        self.product_columns = []
        self.focus_row = 1

    def title_block(self, title:str)->Block:
        cell = Cell(value=title, style=title_style)
        block = Block(
                (cell,),
                self.focus_row, 1,
                )
        self.focus_row += 2
        return block

    def normal_block(self, normal:str)->Block:
        cell = Cell(value=normal, style=normal_style)
        block = Block(
                (cell,),
                self.focus_row, 1
                )
        self.focus_row += 1
        return block

    def reactant_blocks(self, reaction_result:ReactionResult)->Iterator[Block]:
        focus_col=1
        yield Block(
                column=row_header_column,
                anchor_row=self.focus_row,
                anchor_col=focus_col,
                width=ROW_HEADER_WIDTH,
                )

        focus_col = 2
        for reactant in reaction_result.reactants:
            column = substance_column(reactant)
            yield Block(
                    column,
                    anchor_row=self.focus_row,
                    anchor_col=focus_col,
                    width=max(NUMBER_WIDTH, col_width_from_str(column[0].value)),
                    )
            focus_col += 1

        self.reaction_rows.append(self.focus_row)
        self.null_start_cols.append(focus_col)
        self.product_columns.append(substance_column(reaction_result.product))

        self.focus_row += 5


    def product_blocks(self):
        eq_col = max(self.null_start_cols)

        for row, null_start_col, product_column in zip(self.reaction_rows, self.null_start_cols, self.product_columns):
            for shift in range(eq_col - null_start_col):
                yield Block(
                        column=null_column,
                        anchor_row=row,
                        anchor_col=null_start_col+shift
                        )
            
            yield Block(
                    column=eq_column,
                    anchor_row=row,
                    anchor_col=eq_col,
                    width=EQ_WIDTH,
                    )

            yield Block(
                    column=product_column,
                    anchor_row=row,
                    anchor_col=eq_col+1,
                    width=max(NUMBER_WIDTH,col_width_from_str(product_column[0].value))
                    )


    def process(self,bundle:Bundle)->Bundle:
        datalines = bundle.stream

        def stream():
            for dataline in datalines:
                match dataline:
                    case DataLine(LineType.TITLE, title):
                        yield self.title_block(title)
                    case DataLine(LineType.REACTION, reaction_result):
                        yield from self.reactant_blocks(reaction_result)
                    case DataLine(LineType.NORMAL, normal):
                        yield self.normal_block(normal)
            
            if self.reaction_rows:
                yield from self.product_blocks()
        
        return Bundle(context=bundle.context, stream=stream())





from __future__ import annotations
from threading import Thread
from queue import Queue
from openpyxl import Workbook
from openpyxl.styles import Font, Alignment, Border, Side
from openpyxl.utils import get_column_letter
from typing import Iterator
from pathlib import Path
from datetime import datetime
import errno
import os

from ..pipeline_contracts import Context, Bundle, Style, Cell, Block
from ..errors import InputError, OutputError, InitHint


OUTPUT_DIR = "./outputs/"
MIN_COL_WIDTH = 4

def block_producer(q:Queue, blocks: Iterator[Block]):
    try:
        for block in blocks:
            q.put(block)
    except InputError as e:
        print(e)
        os._exit(1)
    except InitHint as e:
        print(e)
        os._exit(0)
    q.put(None)



def block_consumer(q:Queue, ws:Worksheet):
    col_width = {}
    while True:
        block = q.get()
        if block is None:
            q.task_done()
            break
        
        write(block, ws, col_width)
        q.task_done()

    set_col_width(ws,col_width)


def write(block:Block,ws:Worksheet, col_width:dict[int,float]):
    col = block.anchor_col
    if block.width:
        col_width[col] = max(col_width.get(col, MIN_COL_WIDTH) , block.width)
    for offset, cell_desc in enumerate(block.column):
        row = block.anchor_row + offset
        cell = ws.cell(row=row,column=col,value=cell_desc.value)

        cell.number_format = cell_desc.number_format

        if cell_desc.style:
            cell.font = cell_desc.style.font
            cell.alignment = cell_desc.style.alignment
            cell.border = cell_desc.style.border

def set_col_width(ws:Worksheet, col_width:dict[int,float]):
    for col,width in col_width.items():
        ws.column_dimensions[get_column_letter(col)].width = width


def save(wb:Workbook, basename: str):
    if basename is None:
        filename = f"{datetime.now().strftime("%Y-%m-%d")}.xlsx"
    else:
        filename = f"{basename}.xlsx"
    output_path = Path(OUTPUT_DIR) / filename
    try:
        try:
            wb.save(output_path)
            return
        except FileNotFoundError:
            try:
                output_path.parent.mkdir(parents=True, exist_ok=True)
            except Exception:
                raise OutputError(
                    output_path,
                    explanation=f"Failed to create directory '{output_path.parent}'.",
                    hint="Check if you have permissions.",
                ) 
            wb.save(output_path)
    except OSError as e:
        # no enough permission
        if e.errno in {errno.EACCES, errno.EPERM}: 
            raise OutputError( output_path, 
                                explanation="Permission denied. The file may be open in Excel or you lack write permission.", 
                                hint="Close Excel/WPS or check your permissions.", ) 
        # illegal path
        if e.errno in {errno.EINVAL, errno.ENAMETOOLONG, errno.ENOENT}:
            raise OutputError( output_path,
                                hint="This path is invalid on this operating system.",)
        # other error
        raise OutputError(
            output_path,
            explanation=f"OS error ({e.errno}): {e.strerror}",
            hint="Check disk space, filesystem type, or path validity.",
)

class OutputStage:

    def process(self, bundle:Bundle)->Bundle:
        blocks = bundle.stream
        wb = Workbook()
        ws = wb.active
        ws.title = "Recipe"

        q = Queue(maxsize=128)
        t1 = Thread(target=block_producer, args=(q,blocks), daemon=True)
        t2 = Thread(target=block_consumer, args=(q,ws), daemon=True)
        t1.start()
        t2.start()

        t2.join()

        output_basename = bundle.context.title
        save(wb, output_basename)

        return Bundle(bundle.context, None)
        



"""
Balancer – A pipeline-based chemical balancer.
"""

from __future__ import annotations

# Atomic weights
from .atomic_mass import Element

# Core
from .pipeline_base import Stage, Pipeline
from .pipeline_contracts import Context,Bundle,LineType, DataLine, Substance, Reaction, SubstanceResult, ReactionResult, Style, Cell, Block
from .errors import CuteError

# Stages
from .pipeline_stages.input_stage import InputStage
from .pipeline_stages.parse_stage import ParseStage
from .pipeline_stages.compute_stage import ComputeStage
from .pipeline_stages.format_stage import FormatStage
from .pipeline_stages.output_stage import OutputStage

__all__ = [
    # atomic mass
    "Element",
    # core
    "Stage", "Pipeline",
    "Context","Bundle","LineType", "DataLine", "Formula", "Reaction", "SubstanceResult", "ReactionResult", "Style", "Cell", "Block"
    "CuteError",
    # stages
    "InputStage","ParseStage","ComputeStage","FormatStage","OutputStage"
]



from __future__ import annotations
from typing import Iterator, Optional
from enum import Enum, auto
from sympy import Rational
from itertools import chain

from ..pipeline_contracts import Context, Bundle, LineType, DataLine, Substance, Reaction
from ..errors import ParseError

# ========= Utils ========
def toRational(s:str)->Optional[Rational]:
    try:
        return Rational(s)
    except (TypeError,ValueError,ZeroDivisionError):
        raise ParseError(
                explanation=f"Unable to interpret this world as a rational number:",
                hint=s,
                        )

# ========= Parse ========
class State(Enum):
    DEFAULT = auto()
    FORMULA = auto()
    FORMULA_COMPLETE = auto()
    CONTROL = auto()
    INFO = auto()
    MASS = auto()
    BASIS = auto()

class C(Enum):
    UPPER = (str.isupper,)
    LOWER = (str.islower,)
    SPACE = (str.isspace,)
    DIGIT = (lambda c: c.isdigit() or c == '.' or c == '/',)
    ADD = (lambda c:  c == '+',)
    HYPHEN = (lambda c: c == '-',)
    L_ANGLE  = (lambda c: c == '<',)
    R_ANGLE = (lambda c: c == '>',)
    L_BRACKET  = (lambda c: c == '[',)
    R_BRACKET = (lambda c: c == ']',)
    L_PAREN = (lambda c: c == '(',)
    R_PAREN = (lambda c: c == ')',)

    @classmethod
    def detect(cls, c: str)->Optional[C]:
        for t in cls:
            if t.value[0](c):
                return t
        return None


def parse_reaction(raw_reaction:str)-> Optional[Reaction]:
    """ parse reaction reversely """
    reversed_number_cache = []
    reversed_formula_cache = []
    reversed_info_cache = []
    mass_cache = None
    substance_cache = None
    cul_multiplier = Rational(1) 
    multipliers_history = []
    reversed_elements = []
    reversed_stoichs = []
    reversed_reactants = []
    reversed_products = []
    reversed_basis = []
    has_input_mass = False
    product_on_right = None
    all_formulas = set()

    c_it = chain(reversed(raw_reaction),"+ ")
    c = ""
    c_type = None
    last_c = ""
    last_c_type = None
    def c_next():
       nonlocal c, last_c, c_type, last_c_type
       last_c = c
       last_c_type = c_type
       c = next(c_it)
       c_type = C.detect(c)

    state = State.DEFAULT
    try:
        c_next()
        while True:
            # ==== State.DEFAULT ====
            if state is State.DEFAULT:
                if c_type is C.SPACE:
                    c_next()
                elif (c_type is C.DIGIT) or (c_type is C.LOWER) or (c_type is C.UPPER) or (c_type is C.R_PAREN):
                    state = State.FORMULA
                elif c_type is C.R_BRACKET:
                    c_next()
                    state = State.INFO
                elif (c_type is C.R_ANGLE) or (c_type is C.HYPHEN):
                    c_next()
                    state = State.CONTROL
                elif c_type is C.ADD:
                    state = State.FORMULA_COMPLETE
                    c_next()
                else:
                    raise ParseError(
                                     explanation=f"There is a unexpected character '{c}' before '{last_c}'.",
                                     hint="Please check the format of your reaction line.")
            # ==== State.FORMULA ====
            elif state is State.FORMULA:
                if c_type is C.DIGIT:
                    if last_c_type is C.LOWER:
                        if last_c == 'g':
                            state = State.MASS
                        else:
                            raise ParseError(
                                explanation=f"A lowercase letter '{last_c}' appears right after a number '{c}' in the formula.",
                                hint="In a chemical formula, lowercase letters must follow an uppercase letter as part of an element symbol, "
                                    "not a number. Please check your capitalization.")
                    else:
                        reversed_number_cache.append(c)
                        reversed_formula_cache.append(c)
                        c_next()
                elif c_type is C.LOWER:
                    if last_c_type is C.DIGIT:
                        reversed_stoichs.append(Rational(''.join(reversed(reversed_number_cache)))*cul_multiplier)
                        reversed_number_cache = []
                    else:
                        reversed_stoichs.append(cul_multiplier)
                    c_next()
                elif c_type is C.UPPER:
                    if last_c_type is C.LOWER:
                        e = c+last_c
                        reversed_elements.append(e)
                        reversed_formula_cache.append(e)
                    elif last_c_type is C.DIGIT:
                        reversed_elements.append(c)
                        reversed_formula_cache.append(c)
                        reversed_stoichs.append(Rational(''.join(reversed(reversed_number_cache)))*cul_multiplier)
                        reversed_number_cache = []
                    else:
                        reversed_elements.append(c)
                        reversed_formula_cache.append(c)
                        reversed_stoichs.append(cul_multiplier)
                    c_next()
                elif c_type is C.R_PAREN:
                    reversed_formula_cache.append(c)
                    if last_c_type is C.DIGIT:
                        multipliers_history.append(cul_multiplier)
                        cul_multiplier *= Rational(''.join(reversed(reversed_number_cache)))
                        reversed_number_cache = []
                    elif last_c_type is C.LOWER:
                        raise ParseError(
                            explanation=f"A lowercase letter '{last_c}' appears right after right parenthesis '{c}' in the formula.",
                            hint="In a chemical formula, lowercase letters must follow an uppercase letter as part of an element symbol, "
                                "not a right parenthesis. Please check your capitalization.")
                       
                    c_next()
                elif c_type is C.L_PAREN:
                    reversed_formula_cache.append(c)
                    try:
                        cul_multiplier=multipliers_history.pop()
                    except IndexError:
                        raise ParseError(
                            explanation="Detected an opening parenthesis '(' that is never closed.",
                            hint="Make sure all parentheses are balanced.",)
                    if last_c_type is C.LOWER:
                        raise ParseError(
                            explanation=f"A lowercase letter '{last_c}' appears right after left parenthesis '{c}' in the formula.",
                            hint="In a chemical formula, lowercase letters must follow an uppercase letter as part of an element symbol, "
                                "not a left parenthesis. Please check your capitalization.")
                    c_next()
                elif c_type is C.SPACE:
                    if last_c_type is C.LOWER:
                        if last_c == 'g':
                            state = State.MASS
                        else:
                            raise ParseError(
                                explanation=f"A lonely lowercase letter '{last_c}' exists in the formula.",
                                hint=f"In a chemical formula, lowercase letters must follow an uppercase letter as part of an element symbol.")
                    elif last_c_type is C.DIGIT:
                        formula = "".join(reversed(reversed_formula_cache))
                        raise ParseError(
                                hint=f"Unexpected space before '{formula}'. Please write a complete chemical formula here."
                                )
                    else:
                        c_next()
                elif c_type is C.R_BRACKET:
                    if last_c_type is C.LOWER:
                        raise ParseError(
                            explanation=f"A lonely lowercase letter '{last_c}' appears after the bracket '['.",
                            hint=f"In a chemical formula, lowercase letters must follow an uppercase letter as part of an element symbol.")
                    elif last_c_type is C.DIGIT:
                        formula = "".join(reversed(reversed_formula_cache))
                        raise ParseError(
                                hint=f"Unexpected bracket ']' before '{formula}'. Please write a complete chemical formula here."
                                )
                    
                    state = State.INFO
                    c_next()
                else:
                    if last_c_type is C.LOWER:
                        raise ParseError(
                            explanation=f"A lonely lowercase letter '{last_c}' exists in the formula.",
                            hint=f"In a chemical formula, lowercase letters must follow an uppercase letter as part of an element symbol.")
                    elif last_c_type is C.DIGIT:
                        lonely_number = "".join(reversed(reversed_number_cache))
                        raise ParseError(
                            explanation=f"A number '{lonely_number}' appears in a formula without a preceding element or bracket.",
                            hint="Numbers in a chemical formula must follow an element symbol or a closing bracket.")
                    elif multipliers_history:
                        raise ParseError(
                            explanation="Unmatched closing parenthesis ')' detected in the reaction formula.",
                            hint="Each ')' must have a matching '(' before it.")
                    state = state.FORMULA_COMPLETE
                    
                    
            # ==== State.FORMULA_COMPLETE ====
            elif state is State.FORMULA_COMPLETE:
                if reversed_formula_cache:
                    elements_stoich = {}
                    for element,stoich in zip(reversed(reversed_elements),reversed(reversed_stoichs)):
                        elements_stoich[element] = elements_stoich.setdefault(element,Rational(0)) + stoich
                    if product_on_right is None:
                        reversed_reactants.append(
                            Substance(
                                formula="".join(reversed(reversed_formula_cache)),
                                info = "".join(reversed(reversed_info_cache)),
                                elements_stoich=elements_stoich,
                                input_mass=mass_cache,
                            )
                        )
                    else:   
                        reversed_products.append(
                            Substance(
                                formula="".join(reversed(reversed_formula_cache)),
                                info = "".join(reversed(reversed_info_cache)),
                                elements_stoich=elements_stoich,
                                input_mass=mass_cache,
                            )
                        )
                    reversed_formula_cache = []
                    reversed_info_cache = []
                    reversed_elements = []
                    reversed_stoichs = []
                    mass_cache = None
                state = State.DEFAULT
                
            # ==== State.MASS ====
            elif state is State.MASS:
                if (last_c_type is C.LOWER) and (reversed_number_cache):
                    raise ParseError(
                        explanation="Unexpected number detected after the mass unit 'g'.",
                        hint="The mass unit 'g' must appear at the end of the number, "
                            "without any digits following it.\n")
                elif c_type is C.DIGIT:
                    reversed_number_cache.append(c)
                    c_next()
                elif (c_type) is C.SPACE and (not reversed_number_cache):
                    c_next()
                elif c_type is C.LOWER or c_type is C.UPPER:
                    raise ParseError(
                        explanation="A space is missing between the formula and its mass value.",
                        hint="Please add a space before the mass unit to avoid ambiguity.")
                else:
                    if mass_cache is None:
                        mass_cache = toRational(''.join(reversed(reversed_number_cache)))
                        reversed_number_cache = []
                    else:
                        raise ParseError(
                            explanation="Multiple mass values detected for the same substance.",
                            hint="Each substance should have only one mass value.")
                    
                    state = State.FORMULA_COMPLETE

            # ==== State.CONTROL ====
            elif state is State.CONTROL:
                if last_c_type is C.HYPHEN:
                    if c_type is C.L_ANGLE:
                        if product_on_right is None:
                            product_on_right = True
                            c_next()
                            state = State.DEFAULT
                        else:
                            raise ParseError(
                                explanation="You writed more than one reaction arrow(<- or ->).",
                                hint="Each reaction should have exactly one arrow indicating direction. "
                                    "For example: '- A + B -> C' or '- C <- A + B'.",)
                    else:
                        raise ParseError(
                            explanation="A reaction arrow seems incomplete or malformed.",
                            hint="Use a full arrow '<-' or '->' to indicate the reaction direction.\n"
                                "Examples:\n"
                                "  '- A + B -> C'\n"
                                "  '- C <- A + B'",)
                elif last_c_type is C.R_ANGLE:
                    if c_type is C.HYPHEN:
                        if product_on_right is None:
                            product_on_right = False
                            c_next()
                            state = State.DEFAULT
                        else:
                            raise ParseError(
                                explanation="You writed more than one reaction arrow(<- or ->).",
                                hint="Each reaction should have exactly one arrow indicating direction. "
                                    "For example: '- A + B -> C' or '- C <- A + B'.",)
                    else:
                        state = State.BASIS
                # no other branch

            # ==== State.INFO ====
            elif state is State.INFO:
                if c_type is C.L_BRACKET:
                    state = State.FORMULA
                else:
                    reversed_info_cache.append(c)
                c_next()
            # ==== State.BASIS ====
            elif state is State.BASIS:
                if c_type is C.UPPER:
                    if last_c_type is C.LOWER:
                        reversed_basis.append(c+last_c)
                    elif last_c_type is C.UPPER:
                        reversed_basis.append(last_c)
                    c_next()
                elif c_type is C.LOWER:
                    if last_c_type is not C.LOWER:
                        c_next()
                    else:

                        raise ParseError(
                            explanation="Invalid element symbol detected inside angle brackets.",
                            hint="Each element symbol should start with an uppercase letter "
                                "optionally followed by a single lowercase letter.",)

                elif c_type is C.L_ANGLE:
                    c_next()
                    state = State.FORMULA
                elif c_type is C.SPACE:
                    if last_c_type is C.LOWER:
                        raise ParseError(
                                explanation="Invalid element symbol in the basis <...>:",
                                hint=last_c
                                )
                    c_next()
                else:
                    raise ParseError(
                        hint="Illegal character(s) inside angle brackets '<...>' or missing '<' as open.",
                        )
    # ==== StopIteration ====
    except StopIteration:
        if state is not State.DEFAULT:
            raise ParseError(
                explanation="Unmatched closing bracket ']' detected.",
                hint="Every closing bracket must have a corresponding opening one.\n",)
        
        if product_on_right is None:
            raise ParseError(
                explanation="No reaction arrow ('->' or '<-') was found in this line.",
                hint="Each reaction must include one arrow to indicate direction.\n"
                    "Examples:\n"
                    "  '- A + B -> C'\n"
                    "  '- C <- A + B'",)
        elif not product_on_right:
            reversed_products, reversed_reactants = reversed_reactants, reversed_products

        if len(reversed_products) != 1:
            product_formulas = (product.formula for product in reversed(reversed_products))
            raise ParseError(
                explanation="This program is designed for synthesis (combination) reactions with a single product. Try to keep only one of these:",
                hint=", ".join(product_formulas)
                )


        if reversed_basis:
            basis = dict.fromkeys(reversed(reversed_basis))
        else:
            basis = reversed_products[0].elements_stoich
            

        # let's go!
        return Reaction(
                    raw_reaction,
                    reversed_products[0],
                    list(reversed(reversed_reactants)),
                    basis,
                )

# ========== Stage =========
class ParseStage:
    """ Parse the reaction line """
    def process(self, bundle: Bundle)->Bundle:
        datalines = bundle.stream

        def stream():
            for dataline in datalines:
                if dataline.line_type is LineType.REACTION:
                    try:
                        raw_reaction = dataline.data
                        dataline.data = parse_reaction(raw_reaction)
                    except ParseError as e:
                        e = ParseError(
                                raw_reaction,
                                explanation=e.explanation,
                                hint=e.hint,
                                )
                        print(e)
                        continue
                yield dataline

        return Bundle(context=bundle.context, stream=stream())



import argparse
from balancer import Stage, Pipeline
from balancer import InputStage, ParseStage, ComputeStage, FormatStage, OutputStage

def main()->None:
    parser = argparse.ArgumentParser(description="Balancer 2.0 — Chemical equation balancer")
    parser.add_argument(
        "input_file",
        nargs="?",
        default="input.md",
        help="Path to the input Markdown file (default: input.md)",
    )
    args = parser.parse_args()

    Pipeline(
        InputStage(),
        ParseStage(),
        ComputeStage(),
        FormatStage(),
        OutputStage(),
    ).run(args.input_file)

if __name__ == "__main__":
    main()



from __future__ import annotations
from typing import Optional
import textwrap
# ======== No Traceback for CuteError ========
import sys

def handle_exception(exc_type, exc, tb):
    if issubclass(exc_type, CuteError):
        print(exc)
    else:
        sys.__excepthook__(exc_type, exc, tb)

sys.excepthook = handle_exception

# ======== ANSI Styles ========
RESET = "\033[0m"
BOLD = "\033[1m"

YELLOW = "\033[33m"
RED = "\033[31m"
BLUE = "\033[34m"


# ======== Base Classes ========
class CuteError(Exception):
    """ Human-friendly error with a gentle tone and structured wording. """

    def __init__(
        self,
        something: str,
        body: str,
        *,
        begin:str = "😿 Something weng wrong while " ,
        explanation: Optional[str] = None,
        hint: Optional[str] = None,
    ):
        self.begin = begin
        self.something = something
        self.body = body
        self.explanation = explanation
        self.hint = hint
        super().__init__(body)

    def __str__(self) -> str:
        # Apply indentation and coloring
        begin = self.begin
        something = f"{BLUE}{self.something}:{RESET}"
        body = textwrap.indent(f"{RED}{BOLD}{self.body}{RESET}", "   ")
        explanation = (
            f"{self.explanation}" if self.explanation else ""
        )
        hint = (
            textwrap.indent(f"{YELLOW}{self.hint}{RESET}", "   ")
            if self.hint else ""
        )

        return (
            f"--------\n{RESET}"
            f"{begin}{something}\n"
            f"{body}\n"
            f"{explanation}\n"
            f"{hint}\n"
        )

# ======== Specialized Errors ========

class InputError(CuteError):
    def __init__(self, path: str,*,explanation,hint):
        super().__init__(
            something="reading the input file",
            body=path,
            explanation=explanation,
            hint=hint,
        )

class InitHint(CuteError):
    def __init__(self, path: str,*,explanation,hint):
        super().__init__(
            something="The input file is needed",
            body=path,
            begin="🤓 Welcome to use this program! ",
            explanation=explanation,
            hint=hint,
        )


class ParseError(CuteError):
    def __init__(self, raw_reaction=None,*,explanation=None,hint=None):
        super().__init__(
            something="parsing the reaction",
            body=raw_reaction,
            explanation=explanation,
            hint=hint,
        )


class ComputeError(CuteError):
    def __init__(self, raw_reaction=None,*,explanation=None,hint=None):
        super().__init__(
            something="balancing the reaction",
            body=raw_reaction,
            explanation=explanation,
            hint=hint,
        )


class OutputError(CuteError):
    def __init__(self, output_path,*,explanation=None,hint=None):
        super().__init__(
            something="saving the recipe table to",
            body=output_path,
            explanation=explanation,
            hint=hint,
        )
